#include <stdio.h>
#include <curses.h>

#include "world.h"
#include "rule.h"
#include "mode.h"

#include "loop-input.h"

static int loop_input_halfdelay(world *wp);
static int loop_input_before_game(world *wp);
static int loop_input_playing(world *wp);
static int loop_input_after_game(world *wp);

static int
loop_input_halfdelay(world *wp)
{
  int c;

  if (wp == NULL)
  {
    fprintf(stderr, "loop_input_halfdelay: wp is NULL\n");
    return 1;
  }

  halfdelay(1);
  c = getch();

  switch (c)
  {
  case 'Q':
    wp->should_quit = 1;
    /* don't return here */
    break;
  case 'R':
    world_reset(wp);
    break;
  default:
    break;
  }

  /* leave halfdelay mode */
  nocbreak();
  /* enable cbreak again */
  cbreak();

  return 0;
}

static int
loop_input_before_game(world *wp)
{
  int c;

  if (wp == NULL)
  {
    fprintf(stderr, "loop_input_before_game: wp is NULL\n");
    return 1;
  }

  c = getch();

  switch (c)
  {
  case 'Q':
    wp->should_quit = 1;
    return 0;
    break;
  case '5':
  case '.':
    wp->mode = MODE_PLAYING;
    wp->waiting_for_move = 0;
    break;
  case '8':
  case 'k':
  case KEY_UP:
    (wp->option_item)--;
    if (wp->option_item < 0)
      wp->option_item = 2;
    break;
  case '2':
  case 'j':
  case KEY_DOWN:
    (wp->option_item)++;
    if (wp->option_item > 2)
      wp->option_item = 0;
    break;
  case '4':
  case 'h':
  case KEY_LEFT:
    switch (wp->option_item)
    {
    case OPTION_ITEM_BLUE:
      if (wp->player[0] == 0)
        wp->player[0] = 4;
      else
        wp->player[0] = 0;
      break;
    case OPTION_ITEM_RED:
      if (wp->player[1] == 0)
        wp->player[1] = 4;
      else
        wp->player[1] = 0;
      break;
    case OPTION_ITEM_VICTORY_CONDITION:
      (wp->victory_condition)--;
      if (wp->victory_condition < 1)
        wp->victory_condition = 1;
      break;
    default:
      fprintf(stderr, "loop_input_before_game: "
              "decreasing unknown option (%d)\n",
              wp->option_item);
      return 1;
      break;
    }
    break;
  case '6':
  case 'l':
  case KEY_RIGHT:
    switch (wp->option_item)
    {
    case OPTION_ITEM_BLUE:
      if (wp->player[0] == 0)
        wp->player[0] = 4;
      else
        wp->player[0] = 0;
      break;
    case OPTION_ITEM_RED:
      if (wp->player[1] == 0)
        wp->player[1] = 4;
      else
        wp->player[1] = 0;
      break;
    case OPTION_ITEM_VICTORY_CONDITION:
      (wp->victory_condition)++;
      if (wp->victory_condition > 9)
        wp->victory_condition = 9;
      break;
    default:
      fprintf(stderr, "loop_input_before_game: "
              "increasing unknown option (%d)\n",
              wp->option_item);
      return 1;
      break;
    }
    break;
  default:
    break;
  }

  return 0;
}

static int
loop_input_playing(world *wp)
{
  int c;
  int dx;
  int dy;
  int g;

  if (wp == NULL)
  {
    fprintf(stderr, "loop_input_playing: wp is NULL\n");
    return 1;
  }

  if (!(wp->waiting_for_move))
  {
    loop_input_halfdelay(wp);
    return 0;
  }
  if (wp->player[wp->turn_player] > 0)
  {
    return 0;
  }

  c = getch();

  switch (c)
  {
  case 'Q':
    wp->should_quit = 1;
    return 0;
    break;
  case 'R':
    world_reset(wp);
    return 0;
    break;
  case '5':
  case '.':
    if (wp->grid_from < 0)
    {
      g = wp->cursor_y * 9 + wp->cursor_x;
      if ((wp->state[18 + g] >= wp->turn_player * 9 + 0)
          && (wp->state[18 + g] < wp->turn_player * 9 + 9))
      {
        wp->grid_from = g;
      }
    }
    else
    {
      wp->grid_to = wp->cursor_y * 9 + wp->cursor_x;
      if (wp->grid_to == wp->grid_from)
      {
        wp->grid_from = -1;
        wp->grid_to = -1;
      }
      else if (!move_is_legal(wp->state, wp->turn_player,
                              wp->grid_from, wp->grid_to,
                              wp->victory_condition))
      {
        wp->grid_to = -1;
      }
    }
    break;
  case 'c':
    wp->grid_from = -1;
    wp->grid_to = -1;
    break;
  default:
    break;
  }

  dx = 0;
  dy = 0;
  switch (c)
  {
  case '7':
  case 'y':
    dx = -1;
    dy = 1;
    break;
  case '8':
  case 'k':
  case KEY_UP:
    dx = 0;
    dy = 1;
    break;
  case '9':
  case 'u':
    dx = 1;
    dy = 1;
    break;
  case '4':
  case 'h':
  case KEY_LEFT:
    dx = -1;
    dy = 0;
    break;
  case '6':
  case 'l':
  case KEY_RIGHT:
    dx = 1;
    dy = 0;
    break;
  case '1':
  case 'b':
    dx = -1;
    dy = -1;
    break;
  case '2':
  case 'j':
  case KEY_DOWN:
    dx = 0;
    dy = -1;
    break;
  case '3':
  case 'n':
    dx = 1;
    dy = -1;
    break;
  default:
    return 0;
    break;
  }

  wp->cursor_x += dx;
  wp->cursor_y += dy;

  if (wp->cursor_x < 0)
    wp->cursor_x = 0;
  if (wp->cursor_x > 8)
    wp->cursor_x = 8;

  if (wp->cursor_y < 0)
    wp->cursor_y = 0;
  if (wp->cursor_y > 8)
    wp->cursor_y = 8;

  return 0;
}

static int
loop_input_after_game(world *wp)
{
  int c;

  if (wp == NULL)
  {
    fprintf(stderr, "loop_input_after_game: wp is NULL\n");
    return 1;
  }

  c = getch();

  switch (c)
  {
  case 'Q':
    wp->should_quit = 1;
    return 0;
    break;
  case 'R':
    world_reset(wp);
    return 0;
  default:
    break;
  }

  return 0;
}

int
loop_input(world *wp)
{
  if (wp == NULL)
  {
    fprintf(stderr, "loop_input: wp is NULL\n");
    return 1;
  }

  switch (wp->mode)
  {
  case MODE_BEFORE_GAME:
    loop_input_before_game(wp);
    break;
  case MODE_PLAYING:
    loop_input_playing(wp);
    break;
  case MODE_AFTER_GAME:
    loop_input_after_game(wp);
    break;
  default:
    fprintf(stderr, "loop_input: unknown mode (%d)\n", wp->mode);
    return 1;
    break;
  }

  return 0;
}
